@@ -18,7 +18,7 @@ from botocore.config import Config
 from aws_iam_tester import __version__
 
 # Defaults
-DEFAULT_SLEEP_SECONDS=60
+DEFAULT_SLEEP_SECONDS=300
 
 logger = None
 
@@ -31,19 +31,28 @@ boto3_config = Config(
 
 @click.command()
 @click.option('--number-of-runs', '-n', help='Run only a limited number of simulations, and then abort.', type=int, default=-1)
-@click.option('--dry-run/--no-dry-run', '-dr/-ndr', help='Dry run mode will not run the actual policy simulations.', default=False)
-@click.option('--config-file', '-c', help='Config file, default config.yml.', default='config.yml')
-@click.option('--include-system-roles/--no-include-system-roles', '-sr/-nsr', help='Include non-user-assumable roles.', default=False)
-@click.option('--write-to-file/--no-write-to-file', '-w/-nw', help='Write results to file.', default=False)
-@click.option('--debug/--no-debug', '-d/-nd', help='Print debug messages.', default=False)
+@click.option('--dry-run/--no-dry-run', '-dr/-ndr', help='Dry run mode will not run the actual policy simulations. Default: False', default=False)
+@click.option('--config-file', '-c', help='Config file location. Default: config.yml.', default='config.yml')
+@click.option('--include-system-roles/--no-include-system-roles', '-sr/-nsr', help='Include non-user-assumable roles. Default: True', default=True)
+@click.option('--write-to-file/--no-write-to-file', '-w/-nw', help='Write results to file. Default: False', default=False)
+@click.option('--output-location', '-o', help='Output location, either s3 (start with s3://) or locally. Default: ./results', default='./results')
+@click.option('--debug/--no-debug', '-d/-nd', help='Print debug messages. Default: False', default=False)
 @click.version_option(version=__version__)
-def main(number_of_runs, dry_run, config_file, include_system_roles, write_to_file, debug):
+def main(number_of_runs, dry_run, config_file, include_system_roles, write_to_file, output_location, debug):
     setup_logger(
         debug=debug,
         )
     # first get current account id
-    client = boto3.client("sts")
-    account_id = client.get_caller_identity()["Account"]
+    sts_client = boto3.client("sts")
+    account_id = sts_client.get_caller_identity()["Account"]
+
+    # try to get the friendly account name
+    iam_client = boto3.client("iam")
+    try:
+        account_alias = iam_client.list_account_aliases(MaxItems=1)["AccountAliases"][0]
+    except:
+        account_alias = account_id
+    
 
     config, global_exemptions, user_landing_account = read_config(config_file)
 
@@ -59,10 +68,9 @@ def main(number_of_runs, dry_run, config_file, include_system_roles, write_to_fi
     sources.extend(users)
     sources.extend(roles)
 
-    # # for testing
+    # for quick testing
     # sources = [
-    #     # "arn:aws:iam::325855013273:user/test-user-remove-after-20200401",
-    #     "arn:aws:iam::325855013273:role/SHARED_PLATFORM_ENGINEER_ROLE",
+    #     "arn:aws:iam::351902532037:role/powerbi_gateway",
     # ]
 
     results = []
@@ -83,6 +91,11 @@ def main(number_of_runs, dry_run, config_file, include_system_roles, write_to_fi
             exemptions = c["exemptions"]
         exemptions.extend(global_exemptions)
 
+        # check if this test contain a 'limit_to' element
+        limit_to = []
+        if "limit_to" in c:
+            limit_to = c["limit_to"]
+
         # check if the expected result has the expected values
         expected_result = c["expected_result"].lower()
         if expected_result not in ['fail', 'succeed']:
@@ -92,17 +105,29 @@ def main(number_of_runs, dry_run, config_file, include_system_roles, write_to_fi
         logger.debug(f"Run config for actions: {actions} with resources {resources}")
 
         for source in sources:
-            # check if the source is exempt from testing
-            exempt = False
-            for e in exemptions:
-                if re.match(e, source):
-                    exempt = True
-                    if debug:
-                        logger.debug(f"\nSource {source} is exempt from testing using exemption: {e}")
-                    else:
-                        print(colored(f".", "blue"), end="")
-                        sys.stdout.flush()
-                    break
+            if limit_to: # do we have a limit_to?
+                exempt = True
+                for l in limit_to:
+                    if re.match(l, source):
+                        exempt = False
+                        break
+
+                if exempt and debug:
+                    logger.debug(f"\nSource {source} is not included in whitelist")
+                elif exempt:
+                    print(colored(f".", "blue"), end="")
+                    sys.stdout.flush()
+            else: # or is this source exempt from testing
+                exempt = False
+                for e in exemptions:
+                    if re.match(e, source):
+                        exempt = True
+                        if debug:
+                            logger.debug(f"\nSource {source} is exempt from testing using exemption: {e}")
+                        else:
+                            print(colored(f".", "blue"), end="")
+                            sys.stdout.flush()
+                        break
             
             counter += 1
             if number_of_runs > 0 and counter > number_of_runs:
@@ -118,40 +143,42 @@ def main(number_of_runs, dry_run, config_file, include_system_roles, write_to_fi
                     resources=resources,
                 )
 
-                denies = [x for x in evaluation_results if is_denied(x)]
-                allows = [x for x in evaluation_results if not is_denied(x)]
+                if evaluation_results:
+                    denies = [x for x in evaluation_results if is_denied(x)]
+                    allows = [x for x in evaluation_results if not is_denied(x)]
 
-                if expect_failures:
-                    # allows are failures
-                    failures = allows
-                else:
-                    # denies are failures
-                    failures = denies
-
-                success = (len(failures) == 0)
-
-                if success:
-                    if debug:
-                        logger.debug(colored(f"Success: {source}", "green"))
+                    if expect_failures:
+                        # allows are failures
+                        failures = allows
                     else:
-                        print(colored(f".", "green"), end="")
-                        sys.stdout.flush()
-                else:
-                    r = construct_results(
-                        source=source,
-                        expect_failures=expect_failures,
-                        results=failures,
-                        print_results=debug,
-                        )
-                    results.extend(r)
-                    if not debug:
-                        print(colored(f".", "red"), end="")
-                        sys.stdout.flush()
+                        # denies are failures
+                        failures = denies
+
+                    success = (len(failures) == 0)
+
+                    if success:
+                        if debug:
+                            logger.debug(colored(f"Success: {source}", "green"))
+                        else:
+                            print(colored(f".", "green"), end="")
+                            sys.stdout.flush()
+                    else:
+                        r = construct_results(
+                            source=source,
+                            expect_failures=expect_failures,
+                            results=failures,
+                            print_results=debug,
+                            )
+                        results.extend(r)
+                        if not debug:
+                            print(colored(f".", "red"), end="")
+                            sys.stdout.flush()
 
     handle_results(
         results=results,
         write_to_file=write_to_file,
-        account_id=account_id,
+        output_location=output_location,
+        account=account_alias,
     )
 
 def setup_logger(debug):
@@ -236,7 +263,6 @@ def get_iam_users():
 def simulate_policy(source, actions, resources):
     """Simulate a set of actions from a specific principal against a resource"""
     def simulate():
-        client = boto3.client("iam", config=boto3_config)
         response = client.simulate_principal_policy(
             PolicySourceArn=source,
             ActionNames=actions,
@@ -261,16 +287,24 @@ def simulate_policy(source, actions, resources):
         )
         return response["EvaluationResults"]
 
+    client = boto3.client("iam", config=boto3_config)
     try:
         return simulate()
-    except botocore.exceptions.ClientError as ce:
+    except client.exceptions.NoSuchEntityException as nsee:
+        logger.error(f"\nCould not find entity {source} during simulation, has it just been removed?\n{nsee}")
+        # but ignore it
+        pass
+        return None
+    except client.exceptions.ClientError as ce:
         if "throttling" in str(ce).lower():
-            logger.warn(colored("Throttling of API is requested. Sleep for 60 seconds and try again"), "blue")
+            logger.error(colored(f"\nThrottling of API is requested. Sleep for {DEFAULT_SLEEP_SECONDS} seconds and try again\n"), "blue")
             time.sleep(DEFAULT_SLEEP_SECONDS)
             return simulate()
         else:
             raise(ce)
-
+    except Exception as e:
+        logger.error(f"\nError simulating entity {source}\n{e}")
+        raise(e)
 
 def is_denied(evaluationResults):
     return evaluationResults["EvalDecision"] != "allowed"
@@ -305,27 +339,64 @@ def construct_results(source, expect_failures, results, print_results=True):
     return response
 
 
-def handle_results(results, write_to_file, account_id):
+def handle_results(results, write_to_file, output_location, account):
     print("\n\n")
     logger.debug("Handle results")
-    if write_to_file:
-        # first ensure directory exists
-        try:
-            path = f"{os.getcwd()}/results"
-            os.makedirs(path)
-        except OSError as exc:
-            if exc.errno == errno.EEXIST and os.path.isdir(path):
+    if results and write_to_file:
+        timestr = time.strftime("%Y%m%d-%H%M%S")
+        filename = f'results-{account}-{timestr}.json'
+
+        # remove the trailing / if present
+        output_location = output_location.rstrip('/')
+
+        # do we need to write to s3?
+        if output_location.startswith('s3://'):
+            # parse the output location into bucket and key
+            bucket = output_location[5:].split("/")[0]
+            # do we have a prefix?
+            try:
+                prefix = output_location[5:].split("/", 1)[1]
+            except IndexError as ie:
+                prefix = ""
                 pass
-            else:
-                raise
-        try:
-            timestr = time.strftime("%Y%m%d-%H%M%S")
-            filename = f'{path}/results-{account_id}-{timestr}.json' 
-            with open(filename, 'w') as outfile:
-                json.dump(results, outfile, indent=4)
-            logger.info(f'Results for {len(results)} findings are written to {filename}')
-        finally:
-            outfile.close()
+            
+            try:
+                full_filename = f'{prefix}/{filename}'
+                logger.debug(f"Try to write results to s3://{bucket}/{full_filename}")
+                s3_client = boto3.client('s3')
+                s3_client.put_object(
+                    Body=json.dumps(results).encode(),
+                    Bucket=bucket,
+                    Key=full_filename,
+                )
+            except Exception as e:
+                logger.error(f"Could not write results to s3: {e}")
+        else:
+            # first ensure directory exists
+            try:
+                if os.path.isabs(output_location): # is output location absolute?
+                    path = output_location 
+                else:
+                    path = os.path.abspath(output_location)
+                os.makedirs(path)
+            except OSError as exc:
+                if exc.errno == errno.EEXIST and os.path.isdir(path):
+                    pass
+                else:
+                    raise
+            try:
+                full_filename = f'{path}/{filename}' 
+                logger.debug(f"Try to write results to local file system: {full_filename}")
+                with open(full_filename, 'w') as outfile:
+                    json.dump(results, outfile, indent=4)
+            except Exception as e:
+                logger.error(f"Could not write results to file system: {e}")
+            finally:
+                outfile.close()
+
+            logger.info(f'Results for {len(results)} findings are written to {full_filename}')
+    elif not results:
+        logger.info(colored("No findings found!", "green"))
     else:
         logger.info(f"Complete list with {len(results)} failures is printed below:\n")
         print(json.dumps(results, indent=4))
